/*
 * main.cxx
 *
 * Main source for OpenIVR
 *
 * A H.323 IVR application.
 *
 * Copyright (C) 2002 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: main.cxx,v $
 * Revision 1.4  2002/11/10 08:12:42  robertj
 * Moved constants for "well known" ports to better place (OPAL change).
 *
 * Revision 1.3  2002/09/26 03:08:26  robertj
 * Added ability to create certificate file if it does not exist.
 *
 * Revision 1.2  2002/09/26 02:34:53  robertj
 * Changed user input passing to vxml as string, thanks Alexander Kovatch
 *
 * Revision 1.1  2002/08/14 07:09:03  craigs
 * Initial version
 *
 * Revision 1.10  2002/07/29 15:23:23  csoutheren
 * Added festival option
 *
 * Revision 1.9  2002/07/29 13:06:55  csoutheren
 * Updated to latest sources
 *
 */

#include <ptlib.h>
#include <ptlib/pipechan.h>

#include "version.h"
#include "custom.h"
#include "main.h"

PCREATE_PROCESS(OpenIVR);

const WORD DefaultHTTPPort = 42420;
const WORD DefaultH323Port = H323EndPoint::DefaultTcpPort;

const char LogLevelKey[]          = "Log Level";
const char UsernameKey[]          = "Admin User Name";
const char PasswordKey[]          = "Admin Password";
const char HttpPortKey[]          = "HTTP Port";
const char CertificateFileKey[]   = "Certificate";
const char BandwidthKey[]         = "Bandwidth";
const char IPTypeOfServiceKey[]   = "IP Type Of Service";
const char H323PortKey[]          = "H.323 Port";
const char LocalUserNameKey[]     = "Local User Name";
const char VoiceDocumentFileKey[] = "Voice Document File";
const char CallLogFileKey[]       = "Call Log File";
const char AllowG7231Key[]        = "Allow G.723.1";

#define new PNEW

///////////////////////////////////////////////////////////////

OpenIVR::OpenIVR()
  : PSecureHTTPServiceProcess(ProductInfo)
{
}


BOOL OpenIVR::OnStart()
{
  // create a new H.323 endpoint
  endpoint = new MyH323EndPoint(*this);

  // change to the default directory to the one containing the executable
  PDirectory exeDir = GetFile().GetDirectory();

#if defined(_WIN32) && defined(_DEBUG)
  // Special check to aid in using DevStudio for debugging.
  if (exeDir.Find("\\Debug\\") != P_MAX_INDEX)
    exeDir = exeDir.GetParent();
#endif
  exeDir.Change();

  return PSecureHTTPServiceProcess::OnStart();
}


void OpenIVR::OnStop()
{
  PSecureHTTPServiceProcess::OnStop();

  exitMain.Signal();

  delete endpoint;
}

void OpenIVR::OnConfigChanged()
{
}


void OpenIVR::AddUnregisteredText(PHTML &)
{
}


BOOL OpenIVR::Initialise(const char * initMsg)
{
  PWaitAndSignal wait(reloadMutex);

  PConfig cfg("Parameters");

  // Set log level as early as possible
  SetLogLevel((PSystemLog::Level)cfg.GetInteger(LogLevelKey, GetLogLevel()));
  if (GetLogLevel() >= PSystemLog::Warning)
    PTrace::SetLevel(GetLogLevel()-PSystemLog::Warning);
  else
    PTrace::SetLevel(0);

  PString userName = cfg.GetString(UsernameKey, "admin");
  PString password = PHTTPPasswordField::Decrypt(cfg.GetString(PasswordKey, "changeme"));

  // Get the HTTP basic authentication info
  PHTTPSimpleAuth authority(GetName(), userName, password);


  // Create the parameters URL page, and start adding fields to it
  PConfigPage * rsrc = new PConfigPage(*this, "Parameters", "Parameters", authority);

  // HTTP authentication username/password
  // Log level for messages
  rsrc->Add(new PHTTPIntegerField(LogLevelKey,
                                  PSystemLog::Fatal, PSystemLog::NumLogLevels-1,
                                  GetLogLevel(),
                                  "1=Fatal only, 2=Errors, 3=Warnings, 4=Info, 5=Debug"));

  // HTTP authentication username/password
  rsrc->Add(new PHTTPStringField  (UsernameKey, 25, userName));
  rsrc->Add(new PHTTPPasswordField(PasswordKey, 25, password));

  // HTTP Port number to use.
  WORD httpPort = (WORD)cfg.GetInteger(HttpPortKey, DefaultHTTPPort);
  rsrc->Add(new PHTTPIntegerField(HttpPortKey, 1, 65535, httpPort));

#ifdef P_SSL
  // SSL certificate file.
  PFilePath certificateFile = cfg.GetString(CertificateFileKey, "server.pem");
  rsrc->Add(new PHTTPStringField(CertificateFileKey, 50, certificateFile));
  if (!SetServerCertificate(certificateFile, TRUE)) {
    PSYSTEMLOG(Fatal, "OpenIVR\tCould not load certificate \"" << certificateFile << '"');
    return FALSE;
  }
#endif

  // get voice map document
  vxmlDocument = cfg.GetString(VoiceDocumentFileKey, "root.vxml");
  rsrc->Add(new PHTTPStringField(VoiceDocumentFileKey, 50, vxmlDocument));

  // get call log file
  callLogFn = cfg.GetString(CallLogFileKey, "call_log.csv");
  rsrc->Add(new PHTTPStringField(CallLogFileKey, 50, callLogFn));

  // see if G.723.1 is allow
  allowG7231 = cfg.GetBoolean(AllowG7231Key, FALSE);
  rsrc->Add(new PHTTPBooleanField(AllowG7231Key, allowG7231));
  
  // initialise the endpoint
  endpoint->Initialise(cfg, rsrc);

  // Finished the resource to add, generate HTML for it and add to name space
  {
    PServiceHTML html("System Parameters");
    rsrc->BuildHTML(html);
    httpNameSpace.AddResource(rsrc, PHTTPSpace::Overwrite);
  }

  //  create the home page
  if (httpNameSpace.FindResource("welcome.html") == NULL) {
    PHTML html;
    html << PHTML::Title("Welcome to "+GetName())
         << PHTML::Body()
         << GetPageGraphic()
         << PHTML::Paragraph() << "<CENTER>\r\n"

         << PHTML::HotLink("Parameters") << "System Parameters" << PHTML::HotLink()
#if 0
         << PHTML::Paragraph()
         << PHTML::HotLink("Status") << "Status Page" << PHTML::HotLink()
#endif
         << PHTML::Paragraph();

    if (!systemLogFileName && systemLogFileName != "-")
      html << PHTML::HotLink("logfile.txt") << "Full Log File" << PHTML::HotLink()
           << PHTML::BreakLine()
           << PHTML::HotLink("tail_logfile") << "Tail Log File" << PHTML::HotLink()
           << PHTML::Paragraph();
 
    html << PHTML::HRule()
         << GetCopyrightText()
         << PHTML::Body();
    httpNameSpace.AddResource(new PServiceHTTPString("welcome.html", html), PHTTPSpace::Overwrite);
  }

  // create monitoring page
  PString monitorText = "<!--#equival monitorinfo-->";
  httpNameSpace.AddResource(new PServiceHTTPString("monitor.txt", monitorText, "text/plain", authority), PHTTPSpace::Overwrite);

  // set up the HTTP port for listening & start the first HTTP thread
  if (ListenForHTTP(httpPort))
    PSYSTEMLOG(Info, "Opened master socket for HTTP: " << httpListeningSocket->GetPort());
  else {
    PSYSTEMLOG(Fatal, "Cannot run without HTTP port: " << httpListeningSocket->GetErrorText());
    return FALSE;
  }

  PSYSTEMLOG(Info, "Service " << GetName() << ' ' << initMsg);
  return TRUE;
}


void OpenIVR::Main()
{
  while (!exitMain.Wait(PTimeInterval(0, 0, 1))) {
  }
  exitMain.Acknowledge();
}

void OpenIVR::OutputLogLine(const PTime & time, 
                               const PString & addr,
                               const PString & dn,
                               PINDEX duration,
                               const PString & codec)
{ 
  PTime now;
  PStringStream strm;

  strm << time.AsString("yyyy/MM/dd hh:mm:ss,")
       << addr << ','
       << dn << ','
       << duration << ','
       << codec;

  PWaitAndSignal m(logMutex); 

  PTextFile logFile;
  if (!logFile.Open(callLogFn, PFile::ReadWrite))
    return;

  logFile.SetPosition(0, PFile::End);
  if (logFile.GetPosition() == 0)
    logFile << "\"Timestamp\","
               "\"Origin Address\","
               "\"Destination DN\","
               "\"Duration\","
               "\"Codec\""
               "\n";

  logFile << strm << endl;
}


///////////////////////////////////////////////////////////////

MyH323EndPoint::MyH323EndPoint(OpenIVR & _app)
  : app(_app)
{
}

H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference)
{
  return new MyH323Connection(*this, callReference, app);
}

BOOL MyH323EndPoint::Initialise(PConfig & cfg, PConfigPage * rsrc)
{
  // Bandwidth setting
  unsigned bandwidth = cfg.GetInteger(BandwidthKey, GetInitialBandwidth());
  SetInitialBandwidth(bandwidth/100);
  rsrc->Add(new PHTTPIntegerField(BandwidthKey, 100, 10000000, bandwidth));

  // IP TOS byte
  unsigned tos = cfg.GetInteger(IPTypeOfServiceKey, GetRtpIpTypeofService());
  SetRtpIpTypeofService(tos);
  rsrc->Add(new PHTTPIntegerField(IPTypeOfServiceKey, 0, 255, tos));

  // H.323 listening port
  unsigned h323port = cfg.GetInteger(H323PortKey, DefaultH323Port);
  rsrc->Add(new PHTTPIntegerField(H323PortKey, 1, 65535, h323port));

  // Interfaces to listen on
  if (listeners.IsEmpty() || ((H323ListenerTCP&)listeners[0]).GetListenerPort() != h323port) {
    RemoveListener(NULL);
    if (!StartListener(psprintf("*:%u+", h323port))) {
      PSYSTEMLOG(Fatal, "OpenIVR\tCould not open H.323 listener port.");
      return FALSE;
    }
  }

  // Local alias name for H.323 endpoint
  PString name = cfg.GetString(LocalUserNameKey, GetLocalUserName());
  rsrc->Add(new PHTTPStringField(LocalUserNameKey, 30, name));
  if (GetLocalUserName() != name)
    SetLocalUserName(name);

  return TRUE;
}

///////////////////////////////////////////////////////////////

MyH323Connection::MyH323Connection(MyH323EndPoint & _ep, unsigned callReference, OpenIVR & _app)
  : H323Connection(_ep, callReference), ep(_ep), app(_app), 
#ifdef _MSC_VER
#pragma warning(disable:4355)
#endif
    vxml(_app, this)
#ifdef _MSC_VER
#pragma warning(default:4355)
#endif
{
  audioChannelsOpen = 0;
}

MyH323Connection::~MyH323Connection()
{
  PTimeInterval duration = PTime() - callStartTime;
  app.OutputLogLine(callStartTime, 
                    srcAddress,
                    e164Number,
                    (PINDEX)(duration.GetMilliSeconds() + 500) / 1000,
                    codecName);
}

void MyH323Connection::OnSetLocalCapabilities()
{
  // only add G.723.1 if allowed
  if (app.IsG7231Allowed())
    localCapabilities.SetCapability(0, 0, new G7231_File_Capability);

  // add all other capabilities
  localCapabilities.SetCapability(0, 0, new H323_GSM0610Capability);
  localCapabilities.SetCapability(0, 0, new MicrosoftGSMAudioCapability);
  localCapabilities.SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::ALaw, H323_G711Capability::At64k));
  localCapabilities.SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::muLaw, H323_G711Capability::At64k));

  // add user input capabilities
  H323_UserInputCapability::AddAllCapabilities(localCapabilities, 0, 0);

  // do the reordering
  localCapabilities.Reorder(app.GetCodecsToReorder());
  localCapabilities.Remove (app.GetCodecsToRemove());
}


BOOL MyH323Connection::OnReceivedSignalSetup(const H323SignalPDU & setupPDU)
{
  // get incoming e164 number
  setupPDU.GetDestinationE164(e164Number);

  return H323Connection::OnReceivedSignalSetup(setupPDU);
}

H323Connection::AnswerCallResponse
     MyH323Connection::OnAnswerCall(const PString & /* caller */,
                                    const H323SignalPDU & /*setupPDU*/,
                                    H323SignalPDU & /*connectPDU*/)
{
  return AnswerCallNow;
}

BOOL MyH323Connection::OpenAudioChannel(BOOL /*isEncoding*/, 
                                        unsigned /* bufferSize */, 
                                        H323AudioCodec & codec)
{
  BOOL doStart = FALSE;

  // cannot lock this mutex whilst calling StartFunctions as 
  // deadlock may be caused
  {
    PWaitAndSignal mutex(connMutex);
    if (audioChannelsOpen == 0) {
      if (!vxml.Open(!codec.IsRawDataChannelNative())) {
        PTRACE(1, "Cannot open VXML engine");
        return FALSE;
      }
      codecName = codec.GetMediaFormat();
    }

    codec.SetSilenceDetectionMode(H323AudioCodec::NoSilenceDetection);

    // one more channel is open, now
    audioChannelsOpen++;

    doStart = audioChannelsOpen == 2;
  }

  if (!codec.AttachChannel(&vxml, FALSE)) {
    PTRACE(1, "Cannot attach VXML channel to codec");
    return FALSE;
  }

  if (doStart) {

    callStartTime = PTime();

    RTP_Session * rtpSession = GetSession(RTP_Session::DefaultAudioSessionID);
    if (rtpSession == NULL) {
      PTRACE(4, "H323VOICE\tNo RTP session found");
    } else {
      if (!rtpSession->IsDescendant(RTP_UDP::Class())) {
        PTRACE(4, "H323VOICE\tRTP session is not descendant of RTP_UDP");
      } else
        srcAddress = ((RTP_UDP *)rtpSession)->GetRemoteAddress().AsString();
    }

    // set up VXML variables
    vxml.SetVar("session.telephone.dnis", e164Number);
    vxml.SetVar("session.starttime",      callStartTime.AsString("yyyy/MM/dd hh:mm:ss"));
    vxml.SetVar("session.h323.source",    srcAddress);
    vxml.SetVar("session.h323.codec",     codecName);

    PTRACE(2, "OpenIVR\tCall to dn " << e164Number << " from " << srcAddress);

    // execute VXML (if any)
    PTRACE(2, "OpenIVR\tLoading VXML " << app.GetVXMLDocument());
    return vxml.Load(app.GetVXMLDocument());
  }

  return TRUE;
}

void MyH323Connection::OnUserInputString(const PString & value)
{
  vxml.OnUserInput(value);
}

///////////////////////////////////////////////////////////////

IVRVXMLSession::IVRVXMLSession(OpenIVR & _app, MyH323Connection * _conn)
  : OpalVXMLSession(_conn), app(_app)
{
  PTextToSpeech * tts = new PTextToSpeech();
  tts->SetEngine(tts->GetEngines()[0]);
  SetTextToSpeech(tts, TRUE);
}
